#ifndef SINGLE_ACCESS_H
#define SINGLE_ACCESS_H

#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/Analysis/LoopInfo.h"
#include "util.cpp"

using namespace llvm;
using namespace std;

class SingleAccess{
private:
	Value* array;
	Instruction* location;
	BasicBlock* block;
	vector<Value*> indexVector;
	LoopInfo* loopInfo;
	
	vector<Value*> indVarChain;
	vector<Loop*> loopChain;
	vector<vector<int> > matrix;
	

public: 
	SingleAccess(Value* array, Instruction* location, 
		vector<Value*> indexVector, BasicBlock* block, LoopInfo* li);
	Value* getArray(){ return array;};
	Instruction* getLocation(){ return location;};
	BasicBlock* getBasicBlock(){ return block;};
	vector<Value*> getIndexVector(){ return indexVector;};

	void updateIndVarMatrix();
	void printIndVarMatrix();
	
	vector<Value*> getIndVarChain();

	vector<Value*> getReorderedIndexVector(
			vector<vector<int> > transferMatrix, Instruction* inst, BasicBlock* block);

	vector<int> getMatrix();
	
	map<Value*, int> indVarToIterationCount;
	int getIterationCount();
};

vector<int> SingleAccess::getMatrix(){
	vector<int> result;
	for(int i=matrix.size()-1; i>=0; i--){
		result.push_back(matrix[i][0]);
	}
	return result;
}

int SingleAccess::getIterationCount(){
	int result = 1;
	for (map<Value*, int>::iterator i=indVarToIterationCount.begin(); 
			i!=indVarToIterationCount.end(); i++){
		int iterationCount = i->second;
		result *= iterationCount;
	}
	return result;
}

SingleAccess::SingleAccess(Value* arr, Instruction* loc,
                vector<Value*> indexVec, BasicBlock* bb, LoopInfo* li){
        array = arr;
        location = loc;
        indexVector = indexVec;
        block = bb;
	loopInfo = li;
}

vector<Value*> SingleAccess::getIndVarChain(){
	Loop* loop = loopInfo->getLoopFor((const BasicBlock*)block);
	loopChain = (new Util())->getLoopChain(loop);
	vector<Value*> indVarChain;
	for (int i = 0; i < loopChain.size(); i++){
		Loop* loop = loopChain[i];
		Value* indVar = (new Util())->getIndVarFromLoop(loop);
		//errs()<<"INDVAR: "<<*indVar<<"\n";
		if (indVar != NULL)
			indVarChain.push_back(indVar);
	}
	return indVarChain;
}

void SingleAccess::updateIndVarMatrix(){
	indVarChain = getIndVarChain();
	for (int i = 0; i < indexVector.size(); i++){
		vector<int> linearParameters = 
			(new Util())->getLinearParamForIndex(indexVector[i], indVarChain);
		//errs()<<"ACCESS:"<<*(indexVector[i])<<"\n";
		vector<int> newRow;
		for (int j = 0; j < indVarChain.size(); j++){
			//errs()<<"INDVAR="<<*(indVarChain[j])<<" PARAM="<<linearParameters[j]<<"\n";
			newRow.push_back(linearParameters[j]);
		}
		matrix.push_back(newRow);
	}
}

void SingleAccess::printIndVarMatrix(){
	errs()<<"PRINTING MATRIX\n";
	for (int j = 0; j < indVarChain.size(); j++){
		errs()<<"\t\t"<<*(indVarChain[j]);
	}
	errs()<<"\n";
	for (int i = 0; i < indexVector.size(); i++){
		errs()<<"\t";
		for (int j = 0; j < indVarChain.size(); j++){
			errs()<<matrix[i][j]<<"\t";
		}
		errs()<<"  "<<*(indexVector[i]);
		errs()<<"\n";
	}
}

vector<Value*> SingleAccess::getReorderedIndexVector(vector<vector<int> > transferMatrix, Instruction* inst, BasicBlock* block){
	vector<Value*> newIndexVector;
	for (int i = 0; i < transferMatrix.size(); i++){
		vector<int> row = transferMatrix[i];
		Value* value =  BinaryOperator::Create(Instruction::Mul, indexVector[0], 
				ConstantInt::get(Type::getInt32Ty(location->getContext()), row[0], false), "TMP");
		(new Util())->insertInst(block, inst, (Instruction*)value);
		for (int j = 1; j < row.size(); j++){
			Value* tmp = BinaryOperator::Create(Instruction::Mul, indexVector[j], 
					ConstantInt::get(Type::getInt32Ty(location->getContext()), row[j], false), "TMP");
			(new Util())->insertInst(block, inst, (Instruction*)tmp);
			value = BinaryOperator::Create(Instruction::Add, value, tmp, "TMP");
			(new Util())->insertInst(block, inst, (Instruction*)value);
		}
		newIndexVector.push_back(value);
	}
	return newIndexVector;
}

#endif
